@@ -5,9 +5,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2003-2007, LINBIT Information Technologies GmbH.
-   Copyright (C) 2003-2007, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2003-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
+   Copyright (C) 2003-2008, LINBIT Information Technologies GmbH.
+   Copyright (C) 2003-2008, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2003-2008, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -39,32 +39,47 @@ STATIC int _drbd_md_sync_page_io(drbd_dev *mdev,
 				 struct page *page, sector_t sector,
 				 int rw, int size)
 {
-	struct bio *bio = bio_alloc(GFP_NOIO, 1);
-	struct completion event;
+	struct bio *bio;
+	struct drbd_md_io md_io;
 	int ok;
 
+	md_io.mdev = mdev;
+	init_completion(&md_io.event);
+	md_io.error = 0;
+
+	if (rw == WRITE && !test_bit(MD_NO_BARRIER, &mdev->flags))
+	    rw |= (1<<BIO_RW_BARRIER);
+	rw |= (1 << BIO_RW_SYNC);
+
+ retry:
+	bio = bio_alloc(GFP_NOIO, 1);
 	bio->bi_bdev = bdev->md_bdev;
 	bio->bi_sector = sector;
 	ok = (bio_add_page(bio, page, size, 0) == size);
-	if(!ok) goto out;
-	init_completion(&event);
-	bio->bi_private = &event;
+	if (!ok)
+		goto out;
+	bio->bi_private = &md_io;
 	bio->bi_end_io = drbd_md_io_complete;
+	bio->bi_rw = rw;
 
-	if (FAULT_ACTIVE(mdev, (rw & WRITE)? DRBD_FAULT_MD_WR:DRBD_FAULT_MD_RD)) {
-		bio->bi_rw |= rw;
-		bio_endio(bio,bio->bi_size,-EIO);
-	}
-	else {
-#ifdef BIO_RW_SYNC
-		submit_bio(rw | (1 << BIO_RW_SYNC), bio);
-#else
+	dump_internal_bio("Md", mdev, bio, 0);
+
+	if (FAULT_ACTIVE(mdev, (rw & WRITE)? DRBD_FAULT_MD_WR:DRBD_FAULT_MD_RD))
+		bio_endio(bio, -EIO);
+	else
 		submit_bio(rw, bio);
-		drbd_blk_run_queue(bdev_get_queue(bdev->md_bdev));
-#endif
+	wait_for_completion(&md_io.event);
+	ok = bio_flagged(bio, BIO_UPTODATE);
+
+	/* check for unsupported barrier op */
+	if (unlikely(md_io.error == -EOPNOTSUPP && bio_barrier(bio))) {
+		/* Try again with no barrier */
+		WARN("Barriers not supported on meta data device - disabling\n");
+		set_bit(MD_NO_BARRIER,&mdev->flags);
+		rw &= ~(1 << BIO_RW_BARRIER);
+		bio_put(bio);
+		goto retry;
 	}
-	wait_for_completion(&event);
-	ok = test_bit(BIO_UPTODATE, &bio->bi_flags);
  out:
 	bio_put(bio);
 	return ok;
@@ -180,6 +195,8 @@ struct update_al_work {
 	struct lc_element * al_ext;
 	struct completion event;
 	unsigned int enr;
+	/* if old_enr != LC_FREE, write corresponding bitmap sector, too */
+	unsigned int old_enr;
 };
 
 STATIC int w_al_write_transaction(struct Drbd_Conf *, struct drbd_work *, int);
@@ -236,15 +253,6 @@ void drbd_al_begin_io(struct Drbd_Conf *mdev, sector_t sector)
 	wait_event(mdev->al_wait, (al_ext = _al_get(mdev,enr)) );
 
 	if (al_ext->lc_number != enr) {
-		// We have to do write an transaction to AL.
-		unsigned int evicted;
-
-		evicted = al_ext->lc_number;
-
-		if(mdev->state.conn < Connected && evicted != LC_FREE ) {
-			drbd_bm_write_sect(mdev, evicted/AL_EXT_PER_BM_SECT );
-		}
-
 		/* drbd_al_write_transaction(mdev,al_ext,enr);
 		   generic_make_request() are serialized on the
 		   current->bio_tail list now. Therefore we have
@@ -253,6 +261,7 @@ void drbd_al_begin_io(struct Drbd_Conf *mdev, sector_t sector)
 		init_completion(&al_work.event);
 		al_work.al_ext = al_ext;
 		al_work.enr = enr;
+		al_work.old_enr = al_ext->lc_number;
 		al_work.w.cb = w_al_write_transaction;
 		drbd_queue_work_front(&mdev->data.work,&al_work.w);
 		wait_for_completion(&al_work.event);
@@ -302,14 +311,23 @@ void drbd_al_complete_io(struct Drbd_Conf *mdev, sector_t sector)
 STATIC int
 w_al_write_transaction(struct Drbd_Conf *mdev, struct drbd_work *w, int unused)
 {
-	int i,n,mx;
-	unsigned int extent_nr;
+	struct update_al_work *aw = (struct update_al_work*)w;
+	struct lc_element *updated = aw->al_ext;
+	const unsigned int new_enr = aw->enr;
+	const unsigned int evicted = aw->old_enr;
+
 	struct al_transaction* buffer;
 	sector_t sector;
+	int i,n,mx;
+	unsigned int extent_nr;
 	u32 xor_sum=0;
 
-	struct lc_element *updated = ((struct update_al_work*)w)->al_ext;
-	unsigned int new_enr = ((struct update_al_work*)w)->enr;
+	/* do we have to do a bitmap write, first?
+	 * TODO reduce maximum latency:
+	 * submit both bios, then wait for both,
+	 * instead of doing two synchronous sector writes. */
+	if (mdev->state.conn < Connected && evicted != LC_FREE)
+		drbd_bm_write_sect(mdev, evicted/AL_EXT_PER_BM_SECT);
 
 	down(&mdev->md_io_mutex); // protects md_io_buffer, al_tr_cycle, ...
 	buffer = (struct al_transaction*)page_address(mdev->md_io_page);
@@ -522,14 +540,14 @@ struct drbd_atodb_wait {
 	int                error;
 };
 
-STATIC int atodb_endio(struct bio *bio, unsigned int bytes_done, int error)
+STATIC BIO_ENDIO_FN(atodb_endio)
 {
 	struct drbd_atodb_wait *wc = bio->bi_private;
 	struct Drbd_Conf *mdev=wc->mdev;
 	struct page *page;
 	int uptodate = bio_flagged(bio,BIO_UPTODATE);
 
-	if (bio->bi_size) return 1;
+	BIO_ENDIO_FN_START;
 	if (!error && !uptodate) {
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
@@ -538,6 +556,7 @@ STATIC int atodb_endio(struct bio *bio, unsigned int bytes_done, int error)
 		error = -EIO;
 	}
 
+	/* corresponding drbd_io_error is in drbd_al_to_on_disk_bm */
 	drbd_chk_io_error(mdev,error,TRUE);
 	if(error && wc->error == 0) wc->error=error;
 
@@ -551,7 +570,7 @@ STATIC int atodb_endio(struct bio *bio, unsigned int bytes_done, int error)
 	mdev->bm_writ_cnt++;
 	dec_local(mdev);
 
-	return 0;
+	BIO_ENDIO_FN_RETURN;
 }
 
 #define S2W(s)	((s)<<(BM_EXT_SIZE_B-BM_BLOCK_SIZE_B-LN2_BPL))
@@ -671,22 +690,26 @@ void drbd_al_to_on_disk_bm(struct Drbd_Conf *mdev)
 		if (bios[i]==NULL) break;
 		if (FAULT_ACTIVE( mdev, DRBD_FAULT_MD_WR )) {
 			bios[i]->bi_rw = WRITE;
-			bio_endio(bios[i],bios[i]->bi_size,-EIO);
+			bio_endio(bios[i], -EIO);
 		} else {
 			submit_bio(WRITE, bios[i]);
 		}
-
 	}
 
 	drbd_blk_run_queue(bdev_get_queue(mdev->bc->md_bdev));
 
-	// In case we did not submit a single IO do not wait for
-	// them to complete. ( Because we would wait forever here. )
-	//
-	// In case we had IOs and they are already complete, there
-	// is not point in waiting anyways.
-	// Therefore this if() ...
-	if(atomic_read(&wc.count)) wait_for_completion(&wc.io_done);
+	/* In case we did not submit a single IO do not wait for
+	 * them to complete. ( Because we would wait forever here. )
+	 *
+	 * In case we had IOs and they are already complete, there
+	 * is not point in waiting anyways.
+	 * Therefore this if () ... */
+	if (atomic_read(&wc.count)) {
+		wait_for_completion(&wc.io_done);
+		/* flush bitmap to stable storage */
+		if (!test_bit(MD_NO_BARRIER,&mdev->flags))
+			blkdev_issue_flush(mdev->bc->md_bdev, NULL);
+	}
 
 	dec_local(mdev);
 
@@ -699,7 +722,11 @@ void drbd_al_to_on_disk_bm(struct Drbd_Conf *mdev)
 	for(i=0;i<nr_elements;i++) {
 		if(bios[i]==NULL) break;
 		bios[i]->bi_size=0;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
 		atodb_endio(bios[i], MD_HARDSECT, 0);
+#else
+		atodb_endio(bios[i], 0);
+#endif
 	}
 	kfree(bios);
 
@@ -842,6 +869,7 @@ STATIC void drbd_try_clear_on_disk_bm(struct Drbd_Conf *mdev,sector_t sector,
 				     ext->lce.lc_number, ext->rs_left, ext->rs_failed, count);
 				dump_stack();
 				// FIXME brrrgs. should never happen!
+				lc_put(mdev->resync, &ext->lce);
 				drbd_force_state(mdev,NS(conn,Disconnecting));
 				return;
 			}
@@ -904,7 +932,7 @@ STATIC void drbd_try_clear_on_disk_bm(struct Drbd_Conf *mdev,sector_t sector,
 void __drbd_set_in_sync(drbd_dev* mdev, sector_t sector, int size, const char* file, const unsigned int line)
 {
 	/* Is called from worker and receiver context _only_ */
-	unsigned long sbnr,ebnr,lbnr,bnr;
+	unsigned long sbnr,ebnr,lbnr;
 	unsigned long count = 0;
 	sector_t esector, nr_sectors;
 	int wake_up=0;
@@ -947,9 +975,7 @@ void __drbd_set_in_sync(drbd_dev* mdev, sector_t sector, int size, const char* f
 	 * we count rs_{total,left} in bits, not sectors.
 	 */
 	spin_lock_irqsave(&mdev->al_lock,flags);
-	for(bnr=sbnr; bnr <= ebnr; bnr++) {
-		if (drbd_bm_clear_bit(mdev,bnr)) count++;
-	}
+	count = drbd_bm_clear_bits(mdev,sbnr,ebnr);
 	if (count) {
 		// we need the lock for drbd_try_clear_on_disk_bm
 		if(jiffies - mdev->rs_mark_time > HZ*10) {
@@ -985,34 +1011,21 @@ void __drbd_set_out_of_sync(drbd_dev* mdev, sector_t sector, int size, const cha
 	unsigned long sbnr,ebnr,lbnr;
 	sector_t esector, nr_sectors;
 
-	/*  Find codepoints that call set_out_of_sync()
-	unsigned long flags;
-	unsigned int enr;
-	struct bm_extent* ext;
-
-	if(inc_local(mdev)) {
-		enr = BM_SECT_TO_EXT(sector);
-		spin_lock_irqsave(&mdev->al_lock,flags);
-		ext = (struct bm_extent *) lc_find(mdev->resync,enr);
-		if (ext) {
-			WARN("BAD! things will happen, find this.\n");
-			dump_stack();
-		}
-		spin_unlock_irqrestore(&mdev->al_lock,flags);
-		dec_local(mdev);
-	}
-	*/
-
 	if (size <= 0 || (size & 0x1ff) != 0 || size > DRBD_MAX_SEGMENT_SIZE) {
 		ERR("sector: %llus, size: %d\n",(unsigned long long)sector,size);
 		return;
 	}
 
+	if (!inc_local(mdev))
+		return; /* no disk, no metadata, no bitmap to set bits in */
+
 	nr_sectors = drbd_get_capacity(mdev->this_bdev);
 	esector = sector + (size>>9) -1;
 
-	ERR_IF(sector >= nr_sectors) return;
-	ERR_IF(esector >= nr_sectors) esector = (nr_sectors-1);
+	ERR_IF(sector >= nr_sectors)
+		goto out;
+	ERR_IF(esector >= nr_sectors)
+		esector = (nr_sectors-1);
 
 	lbnr = BM_SECT_TO_BIT(nr_sectors-1);
 
@@ -1028,7 +1041,10 @@ void __drbd_set_out_of_sync(drbd_dev* mdev, sector_t sector, int size, const cha
 
 	/* ok, (capacity & 7) != 0 sometimes, but who cares...
 	 * we count rs_{total,left} in bits, not sectors.  */
-	drbd_bm_set_bits_in_irq(mdev,sbnr,ebnr);
+	drbd_bm_set_bits(mdev,sbnr,ebnr);
+
+out:
+	dec_local(mdev);
 }
 
 static inline
@@ -1395,8 +1411,8 @@ int drbd_rs_del_all(drbd_dev* mdev)
 void drbd_rs_failed_io(drbd_dev* mdev, sector_t sector, int size)
 {
 	/* Is called from worker and receiver context _only_ */
-	unsigned long sbnr,ebnr,lbnr,bnr;
-	unsigned long count = 0;
+	unsigned long sbnr, ebnr, lbnr;
+	unsigned long count;
 	sector_t esector, nr_sectors;
 	int wake_up=0;
 
@@ -1437,9 +1453,7 @@ void drbd_rs_failed_io(drbd_dev* mdev, sector_t sector, int size)
 	 * we count rs_{total,left} in bits, not sectors.
 	 */
 	spin_lock_irq(&mdev->al_lock);
-	for(bnr=sbnr; bnr <= ebnr; bnr++) {
-		if (drbd_bm_test_bit(mdev,bnr)>0) count++;
-	}
+	count = drbd_bm_count_bits(mdev,sbnr,ebnr);
 	if (count) {
 		mdev->rs_failed += count;
 
