@@ -5,9 +5,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2001-2007, LINBIT Information Technologies GmbH.
-   Copyright (C) 1999-2007, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
+   Copyright (C) 2001-2008, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2008, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2002-2008, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -45,7 +45,7 @@ char *drbd_m_holder = "Hands off! this is DRBD's meta data device.";
 
 
 // Generate the tag_list to struct functions
-#define PACKET(name, number, fields) \
+#define NL_PACKET(name, number, fields) \
 int name ## _from_tags (drbd_dev *mdev, unsigned short* tags, struct name * arg) \
 { \
 	int tag; \
@@ -65,27 +65,27 @@ int name ## _from_tags (drbd_dev *mdev, unsigned short* tags, struct name * arg)
 	} \
 	return 1; \
 }
-#define INTEGER(pn,pr,member) \
+#define NL_INTEGER(pn,pr,member) \
 	case pn: /* D_ASSERT( tag_type(tag) == TT_INTEGER ); */ \
 		 arg->member = *(int*)(tags); \
 		 break;
-#define INT64(pn,pr,member) \
+#define NL_INT64(pn,pr,member) \
 	case pn: /* D_ASSERT( tag_type(tag) == TT_INT64 ); */ \
 		 arg->member = *(u64*)(tags); \
 		 break;
-#define BIT(pn,pr,member) \
+#define NL_BIT(pn,pr,member) \
 	case pn: /* D_ASSERT( tag_type(tag) == TT_BIT ); */ \
 		 arg->member = *(char*)(tags) ? 1 : 0; \
 		 break;
-#define STRING(pn,pr,member,len) \
+#define NL_STRING(pn,pr,member,len) \
 	case pn: /* D_ASSERT( tag_type(tag) == TT_STRING ); */ \
 		 arg->member ## _len = dlen; \
-		 memcpy(arg->member,tags,dlen); \
+		 memcpy(arg->member,tags,min_t(size_t,dlen,len)); \
 		 break;
 #include "linux/drbd_nl.h"
 
 // Generate the struct to tag_list functions
-#define PACKET(name, number, fields) \
+#define NL_PACKET(name, number, fields) \
 unsigned short* \
 name ## _to_tags (drbd_dev *mdev, struct name * arg, unsigned short* tags) \
 { \
@@ -93,22 +93,22 @@ name ## _to_tags (drbd_dev *mdev, struct name * arg, unsigned short* tags) \
 	return tags; \
 }
 
-#define INTEGER(pn,pr,member) \
+#define NL_INTEGER(pn,pr,member) \
 	*tags++ = pn | pr | TT_INTEGER; \
 	*tags++ = sizeof(int); \
 	*(int*)tags = arg->member; \
 	tags = (unsigned short*)((char*)tags+sizeof(int));
-#define INT64(pn,pr,member) \
+#define NL_INT64(pn,pr,member) \
 	*tags++ = pn | pr | TT_INT64; \
 	*tags++ = sizeof(u64); \
 	*(u64*)tags = arg->member; \
 	tags = (unsigned short*)((char*)tags+sizeof(u64));
-#define BIT(pn,pr,member) \
+#define NL_BIT(pn,pr,member) \
 	*tags++ = pn | pr | TT_BIT; \
 	*tags++ = sizeof(char); \
 	*(char*)tags = arg->member; \
 	tags = (unsigned short*)((char*)tags+sizeof(char));
-#define STRING(pn,pr,member,len) \
+#define NL_STRING(pn,pr,member,len) \
 	*tags++ = pn | pr | TT_STRING; \
 	*tags++ = arg->member ## _len; \
 	memcpy(tags,arg->member, arg->member ## _len); \
@@ -121,12 +121,12 @@ void drbd_nl_send_reply(struct cn_msg *, int);
 
 char *nl_packet_name(int packet_type) {
 // Generate packet type strings
-#define PACKET(name, number, fields) \
+#define NL_PACKET(name, number, fields) \
 	[ P_ ## name ] = # name,
-#define INTEGER Argh!
-#define BIT Argh!
-#define INT64 Argh!
-#define STRING Argh!
+#define NL_INTEGER Argh!
+#define NL_BIT Argh!
+#define NL_INT64 Argh!
+#define NL_STRING Argh!
 
 	static char *nl_tag_name[P_nl_after_last_packet] = {
 #include "linux/drbd_nl.h"
@@ -164,7 +164,7 @@ void nl_trace_reply(void *data) {
 int drbd_khelper(drbd_dev *mdev, char* cmd)
 {
 	char mb[12];
-	char *argv[] = {"/sbin/drbdadm", cmd, mb, NULL };
+	char *argv[] = {usermode_helper, cmd, mb, NULL };
 	static char *envp[] = { "HOME=/",
 				"TERM=linux",
 				"PATH=/sbin:/usr/sbin:/bin:/usr/bin",
@@ -172,8 +172,10 @@ int drbd_khelper(drbd_dev *mdev, char* cmd)
 
 	snprintf(mb,12,"minor-%d",mdev_to_minor(mdev));
 
+	INFO("helper command: %s %s\n",usermode_helper,cmd);
+
 	drbd_bcast_ev_helper(mdev,cmd);
-	return call_usermodehelper("/sbin/drbdadm",argv,envp,1);
+	return call_usermodehelper(usermode_helper,argv,envp,1);
 }
 
 drbd_disks_t drbd_try_outdate_peer(drbd_dev *mdev)
@@ -184,14 +186,14 @@ drbd_disks_t drbd_try_outdate_peer(drbd_dev *mdev)
 
 	D_ASSERT(mdev->state.pdsk == DUnknown);
 
-	fp = DontCare;
-	if(inc_local(mdev)) {
+	if (inc_local_if_state(mdev,UpToDate)) {
 		fp = mdev->bc->dc.fencing;
 		dec_local(mdev);
+	} else {
+		WARN("Not outdating peer, since I am diskless.");
+		return mdev->state.pdsk;
 	}
 
-	D_ASSERT( fp > DontCare );
-
 	if( fp == Stonith ) drbd_request_state(mdev,NS(susp,1));
 
 	r=drbd_khelper(mdev,"outdate-peer");
@@ -221,7 +223,6 @@ drbd_disks_t drbd_try_outdate_peer(drbd_dev *mdev)
 	default:
 		/* The script is broken ... */
 		nps = DUnknown;
-		drbd_request_state(mdev,NS(disk,Outdated));
 		ERR("outdate-peer helper broken, returned %d \n",(r>>8)&0xff);
 		return nps;
 	}
@@ -267,7 +268,7 @@ int drbd_set_role(drbd_dev *mdev, drbd_role_t new_role, int force)
 
 			val.pdsk = nps;
 			mask.pdsk = disk_mask;
-			
+
 			continue;
 		}
 
@@ -299,7 +300,7 @@ int drbd_set_role(drbd_dev *mdev, drbd_role_t new_role, int force)
 		break;
 	}
 
-	if(forced) WARN("Forced to conisder local data as UpToDate!\n");
+	if(forced) WARN("Forced to consider local data as UpToDate!\n");
 
 	drbd_sync_me(mdev);
 
@@ -318,6 +319,10 @@ int drbd_set_role(drbd_dev *mdev, drbd_role_t new_role, int force)
 
 	if (new_role == Secondary) {
 		set_disk_ro(mdev->vdisk, TRUE );
+		if ( inc_local(mdev) ) {
+			mdev->bc->md.uuid[Current] &= ~(u64)1;
+			dec_local(mdev);
+		}
 	} else {
 		if(inc_net(mdev)) {
 			mdev->net_conf->want_lose = 0;
@@ -328,10 +333,14 @@ int drbd_set_role(drbd_dev *mdev, drbd_role_t new_role, int force)
 		mdev->this_bdev->bd_disk = mdev->vdisk;
 		 */
 
-		if ( ( ( mdev->state.conn < Connected ||
-			 mdev->state.pdsk <= Failed ) &&
-		       mdev->bc->md.uuid[Bitmap] == 0) || forced ) {
-			drbd_uuid_new_current(mdev);
+		if ( inc_local(mdev) ) {
+			if ( ( ( mdev->state.conn < Connected ||
+				 mdev->state.pdsk <= Failed ) &&
+			       mdev->bc->md.uuid[Bitmap] == 0) || forced ) {
+				drbd_uuid_new_current(mdev);
+			}
+			mdev->bc->md.uuid[Current] |=  (u64)1;
+			dec_local(mdev);
 		}
 	}
 
@@ -437,9 +446,14 @@ char* ppsize(char* buf, unsigned long long size)
 	return buf;
 }
 
-/* You should call drbd_md_sync() after calling this.
+/**
+ * drbd_determin_dev_size:
+ * Evaluates all constraints and sets our correct device size.
+ * Negative return values indicate errors. 0 and positive values
+ * indicate success.
+ * You should call drbd_md_sync() after calling this function.
  */
-int drbd_determin_dev_size(struct Drbd_Conf* mdev)
+enum determin_dev_size_enum drbd_determin_dev_size(drbd_dev* mdev)
 {
 	sector_t prev_first_sect, prev_size; // previous meta location
 	sector_t la_size;
@@ -447,7 +461,7 @@ int drbd_determin_dev_size(struct Drbd_Conf* mdev)
 	char ppb[10];
 
 	int md_moved, la_size_changed;
-	int rv=0;
+	enum determin_dev_size_enum rv=unchanged;
 
 	wait_event(mdev->al_wait, lc_try_lock(mdev->act_log));
 
@@ -468,7 +482,7 @@ int drbd_determin_dev_size(struct Drbd_Conf* mdev)
 			/* currently there is only one error: ENOMEM! */
 			size = drbd_bm_capacity(mdev)>>1;
 			if (size == 0) {
-				ERR("OUT OF MEMORY! Could not allocate bitmap! Set device size => 0\n");
+				ERR("OUT OF MEMORY! Could not allocate bitmap!\n");
 			} else {
 				/* FIXME this is problematic,
 				 * if we in fact are smaller now! */
@@ -476,7 +490,7 @@ int drbd_determin_dev_size(struct Drbd_Conf* mdev)
 				    "Leaving size unchanged at size = %lu KB\n",
 				    (unsigned long)size);
 			}
-			rv = err;
+			rv = dev_size_error;
 		}
 		// racy, see comments above.
 		drbd_set_my_capacity(mdev,size);
@@ -484,7 +498,7 @@ int drbd_determin_dev_size(struct Drbd_Conf* mdev)
 		INFO("size = %s (%llu KB)\n",ppsize(ppb,size>>1),
 		     (unsigned long long)size>>1);
 	}
-	if (rv < 0) goto out;
+	if (rv == dev_size_error) goto out;
 
 	la_size_changed = (la_size != mdev->bc->md.la_size_sect);
 
@@ -506,6 +520,9 @@ int drbd_determin_dev_size(struct Drbd_Conf* mdev)
 			dec_local(mdev);
 		}
 	}
+
+	if (size > la_size) rv = grew;
+	if (size < la_size) rv = shrunk;
   out:
 	lc_unlock(mdev->act_log);
 
@@ -541,11 +558,11 @@ drbd_new_dev_size(struct Drbd_Conf* mdev, struct drbd_backing_dev *bdev)
 	}
 
 	if(u_size) {
-		if(u_size<<1 > size) {
+		if(u_size > size) {
 			ERR("Requested disk size is too big (%lu > %lu)\n",
-			    (unsigned long)u_size, (unsigned long)size>>1);
+			    (unsigned long)u_size>>1, (unsigned long)size>>1);
 		} else {
-			size = u_size<<1;
+			size = u_size;
 		}
 	}
 
@@ -609,8 +626,8 @@ STATIC int drbd_check_al_size(drbd_dev *mdev)
 
 void drbd_setup_queue_param(drbd_dev *mdev, unsigned int max_seg_s)
 {
-	request_queue_t * const q = mdev->rq_queue;
-	request_queue_t * const b = mdev->bc->backing_bdev->bd_disk->queue;
+	struct request_queue * const q = mdev->rq_queue;
+	struct request_queue * const b = mdev->bc->backing_bdev->bd_disk->queue;
 	//unsigned int old_max_seg_s = q->max_segment_size;
 
 	if (b->merge_bvec_fn && !mdev->bc->dc.use_bmbv)
@@ -792,7 +809,7 @@ STATIC int drbd_nl_disk_conf(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 		goto release_bdev2_fail;
 	}
 
-	if ((drbd_get_capacity(nbc->backing_bdev)>>1) < nbc->dc.disk_size) {
+	if ((drbd_get_capacity(nbc->backing_bdev)) < nbc->dc.disk_size) {
 		retcode = LDDeviceTooSmall;
 		goto release_bdev2_fail;
 	}
@@ -833,10 +850,13 @@ STATIC int drbd_nl_disk_conf(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 		goto release_bdev2_fail;
 	}
 
+	nbc->known_size = drbd_get_capacity(nbc->backing_bdev);
+
 	if((retcode = drbd_request_state(mdev,NS(disk,Attaching))) < SS_Success ) {
 		goto release_bdev2_fail;
 	}
 
+	drbd_thread_start(&mdev->worker);
 	drbd_md_set_sector_offsets(mdev,nbc);
 
 	retcode = drbd_md_read(mdev,nbc);
@@ -911,7 +931,10 @@ STATIC int drbd_nl_disk_conf(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 	}
 
 	drbd_bm_lock(mdev); // racy...
-	drbd_determin_dev_size(mdev);
+	if (drbd_determin_dev_size(mdev) == dev_size_error) {
+		retcode = VMallocFailed;
+		goto unlock_bm;
+	}
 
 	if (drbd_md_test_flag(mdev->bc,MDF_FullSync)) {
 		INFO("Assuming that all blocks are out of sync (aka FullSync)\n");
@@ -979,13 +1002,24 @@ STATIC int drbd_nl_disk_conf(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 	rv = _drbd_set_state(mdev, ns, ChgStateVerbose);
 	ns = mdev->state;
 	spin_unlock_irq(&mdev->req_lock);
-	if (rv==SS_Success) after_state_ch(mdev,os,ns,ChgStateVerbose);
 
 	if (rv < SS_Success) {
 		goto unlock_bm;
 	}
 
 	drbd_bm_unlock(mdev);
+
+	if(inc_local_if_state(mdev,Attaching)) {
+		if(mdev->state.role == Primary) mdev->bc->md.uuid[Current] |=  (u64)1;
+		else                            mdev->bc->md.uuid[Current] &= ~(u64)1;
+		dec_local(mdev);
+	}
+
+	/* Reset the "barriers don't work" bits here, then force meta data to
+	 * be written, to ensure we determine if barriers are supported. */
+	clear_bit(LL_DEV_NO_FLUSH,&mdev->flags);
+	clear_bit(MD_NO_BARRIER,&mdev->flags);
+	drbd_md_mark_dirty(mdev);
 	drbd_md_sync(mdev);
 
 	reply->ret_code = retcode;
@@ -1196,6 +1230,9 @@ FIXME LGE
 
 	retcode = drbd_request_state(mdev,NS(conn,Unconnected));
 
+	if (retcode >= SS_Success)
+		drbd_thread_start(&mdev->worker);
+
 	reply->ret_code = retcode;
 	return 0;
 
@@ -1226,11 +1263,12 @@ STATIC int drbd_nl_disconnect(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 		// The peer probabely wants to see us outdated.
 		retcode = _drbd_request_state(mdev,NS2(conn,Disconnecting,
 						       disk,Outdated),0);
-		if( retcode == SS_IsDiskLess ) {
+		if (retcode == SS_IsDiskLess || retcode == SS_LowerThanOutdated) {
 			// We are diskless and our peer wants to outdate us.
 			// So, simply go away, and let the peer try to
 			// outdate us with its 'outdate-peer' handler later.
-			retcode = drbd_request_state(mdev,NS(conn,StandAlone));
+			drbd_force_state(mdev, NS(conn, Disconnecting));
+			retcode = SS_Success;
 		}
 	}
 
@@ -1250,11 +1288,29 @@ STATIC int drbd_nl_disconnect(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 	return 0;
 }
 
+void resync_after_online_grow(drbd_dev *mdev)
+{
+	int iass; /* I am sync source */
+
+	INFO("Resync of new storage after online grow\n");
+	if (mdev->state.role != mdev->state.peer)
+		iass = (mdev->state.role == Primary);
+	else
+		iass = test_bit(DISCARD_CONCURRENT,&mdev->flags);
+
+	if (iass)
+		drbd_start_resync(mdev,SyncSource);
+	else
+		drbd_request_state(mdev,NS(conn,WFSyncUUID));
+}
+
 STATIC int drbd_nl_resize(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 			  struct drbd_nl_cfg_reply *reply)
 {
 	struct resize rs;
 	int retcode=NoError;
+	int ldsc = 0; /* local disk size changed */
+	enum determin_dev_size_enum dd;
 
 	memset(&rs, 0, sizeof(struct resize));
 	if (!resize_from_tags(mdev,nlp->tag_list,&rs)) {
@@ -1278,15 +1334,27 @@ STATIC int drbd_nl_resize(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 		goto fail;
 	}
 
+	if (mdev->bc->known_size != drbd_get_capacity(mdev->bc->backing_bdev)) {
+		mdev->bc->known_size = drbd_get_capacity(mdev->bc->backing_bdev);
+		ldsc = 1;
+	}
+
 	mdev->bc->dc.disk_size = (sector_t)rs.resize_size;
 	drbd_bm_lock(mdev);
-	drbd_determin_dev_size(mdev);
+	dd = drbd_determin_dev_size(mdev);
 	drbd_md_sync(mdev);
 	drbd_bm_unlock(mdev);
 	dec_local(mdev);
-	if (mdev->state.conn == Connected) {
-		drbd_send_uuids(mdev); // to start sync...
+	if (dd == dev_size_error) {
+		retcode = VMallocFailed;
+		goto fail;
+	}
+
+	if (mdev->state.conn == Connected && ( dd != unchanged || ldsc) ) {
+		drbd_send_uuids(mdev);
 		drbd_send_sizes(mdev);
+		if (dd == grew)
+			resync_after_online_grow(mdev);
 	}
 
  fail:
@@ -1423,29 +1491,7 @@ STATIC int drbd_nl_resume_io(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 STATIC int drbd_nl_outdate(drbd_dev *mdev, struct drbd_nl_cfg_req *nlp,
 			   struct drbd_nl_cfg_reply *reply)
 {
-	int retcode;
-	drbd_state_t os,ns;
-
-	spin_lock_irq(&mdev->req_lock);
-	os = mdev->state;
-	if( mdev->state.disk < Outdated ) {
-		retcode = -999;
-	} else {
-		retcode = _drbd_set_state(_NS(mdev,disk,Outdated),ChgStateVerbose);
-	}
-	ns = mdev->state;
-	spin_unlock_irq(&mdev->req_lock);
-	if (retcode==SS_Success) after_state_ch(mdev,os,ns, ChgStateVerbose);
-
-	if( retcode == -999 ) {
-		retcode = DiskLowerThanOutdated;
-		goto fail;
-	}
-
-	drbd_md_sync(mdev);
-
- fail:
-	reply->ret_code = retcode;
+	reply->ret_code = drbd_request_state(mdev,NS(disk,Outdated));
 	return 0;
 }
 
@@ -1660,7 +1706,7 @@ void drbd_connector_callback(void *data)
 
 atomic_t drbd_nl_seq = ATOMIC_INIT(2); // two.
 
-void drbd_bcast_state(drbd_dev *mdev)
+void drbd_bcast_state(drbd_dev *mdev, drbd_state_t state)
 {
 	char buffer[sizeof(struct cn_msg)+
 		    sizeof(struct drbd_nl_cfg_reply)+
@@ -1672,7 +1718,7 @@ void drbd_bcast_state(drbd_dev *mdev)
 
 	// WARN("drbd_bcast_state() got called\n");
 
-	tl = get_state_to_tags(mdev,(struct get_state*)&mdev->state,tl);
+	tl = get_state_to_tags(mdev,(struct get_state*)&state,tl);
 	*tl++ = TT_END; /* Close the tag list */
 
 	cn_reply->id.idx = CN_IDX_DRBD;
@@ -1740,27 +1786,19 @@ void drbd_bcast_sync_progress(drbd_dev *mdev)
 	struct cn_msg *cn_reply = (struct cn_msg *) buffer;
 	struct drbd_nl_cfg_reply* reply = (struct drbd_nl_cfg_reply*)cn_reply->data;
 	unsigned short *tl = reply->tag_list;
-	int res;
 	unsigned long rs_left;
+	unsigned int res;
 
-	if (inc_local(mdev)) {
-		typecheck(unsigned long, mdev->rs_total);
+	/* no local ref, no bitmap, no syncer progress, no broadcast. */
+	if (!inc_local(mdev))
+		return;
+	drbd_get_syncer_progress(mdev, &rs_left, &res);
+	dec_local(mdev);
 
-		rs_left = drbd_bm_total_weight(mdev) - mdev->rs_failed;
-		if (rs_left > mdev->rs_total) {
-			ERR("logic bug? rs_left=%lu > rs_total=%lu (rs_failed %lu)\n",
-					rs_left, mdev->rs_total, mdev->rs_failed);
-			res = 1000;
-		} else {
-			res = (rs_left >> 10)*1000/((mdev->rs_total >> 10) + 1);
-		}
-		dec_local(mdev);
-		res = 1000L - res;
-		*tl++ = T_sync_progress;
-		*tl++ = sizeof(int);
-		memcpy(tl, &res, sizeof(int));
-		tl=(unsigned short*)((char*)tl + sizeof(int));
-	}
+	*tl++ = T_sync_progress;
+	*tl++ = sizeof(int);
+	memcpy(tl, &res, sizeof(int));
+	tl=(unsigned short*)((char*)tl + sizeof(int));
 	*tl++ = TT_END; /* Close the tag list */
 
 	cn_reply->id.idx = CN_IDX_DRBD;
