@@ -5,9 +5,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2001-2007, LINBIT Information Technologies GmbH.
-   Copyright (C) 1999-2007, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
+   Copyright (C) 2001-2008, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2008, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2002-2008, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -62,20 +62,28 @@
 /* used for synchronous meta data and bitmap IO
  * submitted by drbd_md_sync_page_io()
  */
-int drbd_md_io_complete(struct bio *bio, unsigned int bytes_done, int error)
+BIO_ENDIO_FN(drbd_md_io_complete)
 {
-	if (bio->bi_size) return 1;
+	struct drbd_md_io *md_io;
+
+	BIO_ENDIO_FN_START;
 	/* error parameter ignored:
 	 * drbd_md_sync_page_io explicitly tests bio_uptodate(bio); */
 
-	complete((struct completion*)bio->bi_private);
-	return 0;
+	md_io = (struct drbd_md_io *)bio->bi_private;
+
+	md_io->error = error;
+
+	dump_internal_bio("Md", md_io->mdev, bio, 1);
+
+	complete(&md_io->event);
+	BIO_ENDIO_FN_RETURN;
 }
 
 /* reads on behalf of the partner,
  * "submitted" by the receiver
  */
-int drbd_endio_read_sec(struct bio *bio, unsigned int bytes_done, int error)
+BIO_ENDIO_FN(drbd_endio_read_sec)
 {
 	unsigned long flags=0;
 	struct Tl_epoch_entry *e=NULL;
@@ -88,7 +96,7 @@ int drbd_endio_read_sec(struct bio *bio, unsigned int bytes_done, int error)
 	/* We are called each time a part of the bio is finished, but
 	 * we are only interested when the whole bio is finished, therefore
 	 * return as long as bio->bio_size is positive.  */
-	if (bio->bi_size) return 1;
+	BIO_ENDIO_FN_START;
 	if (!error && !uptodate) {
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
@@ -99,6 +107,8 @@ int drbd_endio_read_sec(struct bio *bio, unsigned int bytes_done, int error)
 
 	D_ASSERT(e->block_id != ID_VACANT);
 
+	dump_internal_bio("Sec", mdev, bio, 1);
+
 	spin_lock_irqsave(&mdev->req_lock,flags);
 	mdev->read_cnt += e->size >> 9;
 	list_del(&e->w.list);
@@ -113,13 +123,13 @@ int drbd_endio_read_sec(struct bio *bio, unsigned int bytes_done, int error)
 	       INFO("Moved EE (READ) to worker sec=%llus size=%u ee=%p\n",
 		    (unsigned long long)e->sector,e->size,e);
 	       );
-	return 0;
+	BIO_ENDIO_FN_RETURN;
 }
 
 /* writes on behalf of the partner, or resync writes,
  * "submitted" by the receiver.
  */
-int drbd_endio_write_sec(struct bio *bio, unsigned int bytes_done, int error)
+BIO_ENDIO_FN(drbd_endio_write_sec)
 {
 	unsigned long flags=0;
 	struct Tl_epoch_entry *e=NULL;
@@ -133,8 +143,7 @@ int drbd_endio_write_sec(struct bio *bio, unsigned int bytes_done, int error)
 	e = bio->bi_private;
 	mdev = e->mdev;
 
-	// see above
-	if (bio->bi_size) return 1;
+	BIO_ENDIO_FN_START;
 	if (!error && !uptodate) {
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
@@ -145,6 +154,8 @@ int drbd_endio_write_sec(struct bio *bio, unsigned int bytes_done, int error)
 
 	D_ASSERT(e->block_id != ID_VACANT);
 
+	dump_internal_bio("Sec", mdev, bio, 1);
+
 	spin_lock_irqsave(&mdev->req_lock,flags);
 	mdev->writ_cnt += e->size >> 9;
 	is_syncer_req = is_syncer_block_id(e->block_id);
@@ -187,12 +198,12 @@ int drbd_endio_write_sec(struct bio *bio, unsigned int bytes_done, int error)
 	wake_asender(mdev);
 	dec_local(mdev);
 
-	return 0;
+	BIO_ENDIO_FN_RETURN;
 }
 
 /* read, readA or write requests on Primary comming from drbd_make_request
  */
-int drbd_endio_pri(struct bio *bio, unsigned int bytes_done, int error)
+BIO_ENDIO_FN(drbd_endio_pri)
 {
 	unsigned long flags;
 	drbd_request_t *req=bio->bi_private;
@@ -200,8 +211,8 @@ int drbd_endio_pri(struct bio *bio, unsigned int bytes_done, int error)
 	drbd_req_event_t what;
 	int uptodate = bio_flagged(bio, BIO_UPTODATE);
 
-	// see above
-	if (bio->bi_size) return 1;
+	BIO_ENDIO_FN_START;
+
 	if (!error && !uptodate) {
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
@@ -210,6 +221,8 @@ int drbd_endio_pri(struct bio *bio, unsigned int bytes_done, int error)
 		error = -EIO;
 	}
 
+	dump_internal_bio("Pri", mdev, bio, 1);
+
 	/* to avoid recursion in _req_mod */
 	what = error
 	       ? (bio_data_dir(bio) == WRITE)
@@ -219,7 +232,7 @@ int drbd_endio_pri(struct bio *bio, unsigned int bytes_done, int error)
 	spin_lock_irqsave(&mdev->req_lock,flags);
 	_req_mod(req, what, error);
 	spin_unlock_irqrestore(&mdev->req_lock,flags);
-	return 0;
+	BIO_ENDIO_FN_RETURN;
 }
 
 int w_io_error(drbd_dev* mdev, struct drbd_work* w,int cancel)
@@ -240,8 +253,6 @@ int w_io_error(drbd_dev* mdev, struct drbd_work* w,int cancel)
 	 * when it is done and had a local write error, see comments there */
 	drbd_req_free(req);
 
-	if(unlikely(cancel)) return 1;
-
 	ok = drbd_io_error(mdev, FALSE);
 	if(unlikely(!ok)) ERR("Sending in w_io_error() failed\n");
 	return ok;
@@ -251,6 +262,11 @@ int w_read_retry_remote(drbd_dev* mdev, struct drbd_work* w,int cancel)
 {
 	drbd_request_t *req = (drbd_request_t*)w;
 
+	/* FIXME this is ugly. we should not detach for read io-error,
+	 * but try to WRITE the DataReply to the failed location,
+	 * to give the disk the chance to relocate that block */
+	drbd_io_error(mdev,FALSE); /* tries to schedule a detach and notifies peer */
+
 	spin_lock_irq(&mdev->req_lock);
 	if ( cancel ||
 	     mdev->state.conn < Connected ||
@@ -263,10 +279,6 @@ int w_read_retry_remote(drbd_dev* mdev, struct drbd_work* w,int cancel)
 	}
 	spin_unlock_irq(&mdev->req_lock);
 
-	/* FIXME this is ugly. we should not detach for read io-error,
-	 * but try to WRITE the DataReply to the failed location,
-	 * to give the disk the chance to relocate that block */
-	drbd_io_error(mdev,FALSE); /* tries to schedule a detach and notifies peer */
 	return w_send_read_req(mdev,w,0);
 }
 
@@ -325,13 +337,6 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 		ERR("%s in w_make_resync_request\n", conns_to_name(mdev->state.conn));
 	}
 
-        number = SLEEP_TIME*mdev->sync_conf.rate / ((BM_BLOCK_SIZE/1024)*HZ);
-
-	if (atomic_read(&mdev->rs_pending_cnt)>number) {
-		goto requeue;
-	}
-	number -= atomic_read(&mdev->rs_pending_cnt);
-
 	if(!inc_local(mdev)) {
 		/* Since we only need to access mdev->rsync a
 		   inc_local_if_state(mdev,Failed) would be sufficient, but
@@ -341,6 +346,14 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 		mdev->resync_work.cb = w_resync_inactive;
 		return 1;
 	}
+	/* All goto requeses have to happend after this block: inc_local() */
+
+	number = SLEEP_TIME*mdev->sync_conf.rate / ((BM_BLOCK_SIZE/1024)*HZ);
+
+	if (atomic_read(&mdev->rs_pending_cnt)>number) {
+		goto requeue;
+	}
+	number -= atomic_read(&mdev->rs_pending_cnt);
 
 	for(i=0;i<number;i++) {
 
@@ -796,10 +809,10 @@ STATIC int _drbd_pause_after(drbd_dev *mdev)
 	int i, rv = 0;
 
 	for (i=0; i < minor_count; i++) {
-		if( !(odev = minor_to_mdev(i)) ) continue;
+		if (!(odev = minor_to_mdev(i)) ||
+		    (odev->state.conn == StandAlone && odev->state.disk == Diskless) ) continue;
 		if (! _drbd_may_sync_now(odev)) {
-			rv |= ( _drbd_set_state(_NS(odev,aftr_isp,1),
-						ChgStateHard|ScheduleAfter)
+			rv |= ( _drbd_set_state(_NS(odev,aftr_isp,1),ChgStateHard)
 				!= SS_NothingToDo ) ;
 		}
 	}
@@ -822,8 +835,7 @@ STATIC int _drbd_resume_next(drbd_dev *mdev)
 		if( !(odev = minor_to_mdev(i)) ) continue;
 		if ( odev->state.aftr_isp ) {
 			if (_drbd_may_sync_now(odev)) {
-				rv |= ( _drbd_set_state(_NS(odev,aftr_isp,0),
-							ChgStateHard|ScheduleAfter)
+				rv |= ( _drbd_set_state(_NS(odev,aftr_isp,0),ChgStateHard)
 					!= SS_NothingToDo ) ;
 			}
 		}
@@ -870,7 +882,7 @@ void drbd_alter_sa(drbd_dev *mdev, int na)
  */
 void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 {
-	drbd_state_t os,ns;
+	drbd_state_t ns;
 	int r=0;
 
 	MTRACE(TraceTypeResync, TraceLvlSummary,
@@ -896,7 +908,7 @@ void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 	}
 
 	drbd_global_lock();
-	ns = os = mdev->state;
+	ns = mdev->state;
 
 	ns.aftr_isp = !_drbd_may_sync_now(mdev);
 
@@ -923,8 +935,6 @@ void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 	drbd_global_unlock();
 
 	if ( r == SS_Success ) {
-		after_state_ch(mdev,os,ns,ChgStateVerbose);
-
 		INFO("Began resync as %s (will sync %lu KB [%lu bits set]).\n",
 		     conns_to_name(ns.conn),
 		     (unsigned long) mdev->rs_total << (BM_BLOCK_SIZE_B-10),
@@ -1038,7 +1048,6 @@ int drbd_worker(struct Drbd_thread *thi)
 
 	D_ASSERT( mdev->state.disk == Diskless && mdev->state.conn == StandAlone );
 	drbd_mdev_cleanup(mdev);
-	module_put(THIS_MODULE);
 
 	INFO("worker terminated\n");
 
