@@ -6,7 +6,7 @@
 set -e
 sorry() {
 	cat <<___
-	Sorry, automagic adjustment of drdb_config.h failed.
+	Sorry, automagic adjustment of drbd_config.h failed.
 	For well known 2.6. kernels, no adjustment to the shipped drbd_config is necessary.
 	You need to verify it yourself.
 ___
@@ -63,7 +63,9 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   # it simply does not work always...
   for f in $O/include/asm/atomic.h \
     $O/include/asm/arch/atomic.h \
-    $O/include2/asm/atomic.h
+    $O/include2/asm/atomic.h \
+    $O/include/asm/atomic_32.h \
+    $O/include/asm/arch/atomic_32.h
   do
     if grep_q "atomic_add_return" $f; then
       have_atomic_add=1
@@ -90,6 +92,17 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   else
     need_backport_of_kzalloc=1
   fi
+  if test -e $KDIR/include/linux/scatterlist.h ; then
+    have_linux_scatterlist_h=1
+    if grep_q "sg_set_buf" $KDIR/include/linux/scatterlist.h ; then
+      need_sg_set_buf=0
+    else
+      need_sg_set_buf=1
+    fi
+  else
+    have_linux_scatterlist_h=0
+    need_sg_set_buf=1
+  fi
 else
     # not a 2.6. kernel. just leave it alone...
     exit 0
@@ -113,6 +126,10 @@ perl -pe "
   { ( $have_nl_dst_groups ? '' : '//' ) . \$1}e;
  s{.*(#define NEED_BACKPORT_OF_KZALLOC.*)}
   { ( $need_backport_of_kzalloc ? '' : '//' ) . \$1}e;
+ s{.*(#define NEED_SG_SET_BUF.*)}
+  { ( $need_sg_set_buf ? '' : '//' ) . \$1}e;
+ s{.*(#define HAVE_LINUX_SCATTERLIST_H.*)}
+  { ( $have_linux_scatterlist_h ? '' : '//' ) . \$1}e;
  " \
 	  < ./linux/drbd_config.h \
 	  > ./linux/drbd_config.h.new
