@@ -3,9 +3,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2002-2007, LINBIT Information Technologies GmbH.
-   Copyright (C) 2002-2007, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
+   Copyright (C) 2002-2008, LINBIT Information Technologies GmbH.
+   Copyright (C) 2002-2008, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2002-2008, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -105,11 +105,14 @@ static int sh_ip(struct d_resource* ,const char* );
 static int sh_ll_dev(struct d_resource* ,const char* );
 static int sh_md_dev(struct d_resource* ,const char* );
 static int sh_md_idx(struct d_resource* ,const char* );
+static int sh_b_pri(struct d_resource* ,const char* );
 static int admm_generic(struct d_resource* ,const char* );
 static int adm_khelper(struct d_resource* ,const char* );
 static int adm_generic_b(struct d_resource* ,const char* );
 static int hidden_cmds(struct d_resource* ,const char* );
 
+static char* get_opt_val(struct d_option*,const char*,char*);
+
 static struct ifreq* get_ifreq();
 
 char ss_buffer[255];
@@ -130,6 +133,7 @@ int verbose;
 int do_verify_ips;
 char* drbdsetup;
 char* drbdmeta;
+char* sh_varname = NULL;
 char* setup_opts[10];
 int soi=0;
 volatile int alarm_raised;
@@ -189,6 +193,7 @@ struct option admopt[] = {
   { "config-file",  required_argument,0, 'c' },
   { "drbdsetup",    required_argument,0, 's' },
   { "drbdmeta",     required_argument,0, 'm' },
+  { "sh-varname",   required_argument,0, 'n' },
   { 0,              0,                0, 0   }
 };
 
@@ -200,8 +205,8 @@ struct adm_cmd cmds[] = {
   { "disconnect",        adm_generic_s, 1,1,0 },
   { "up",                adm_up,        1,1,1 },
   { "down",              adm_generic_s, 1,1,0 },
-  { "primary",           adm_generic_s, 1,1,0 },
-  { "secondary",         adm_generic_s, 1,1,0 },
+  { "primary",           adm_generic_l, 1,1,0 },
+  { "secondary",         adm_generic_l, 1,1,0 },
   { "invalidate",        adm_generic_b, 1,1,0 },
   { "invalidate-remote", adm_generic_l, 1,1,1 },
   { "outdate",           adm_generic_b, 1,1,0 },
@@ -212,8 +217,8 @@ struct adm_cmd cmds[] = {
   { "adjust",            adm_adjust,    1,1,1 },
   { "wait-connect",      adm_wait_c,    1,1,1 },
   { "state",             adm_generic_s, 1,1,0 },
-  { "cstate",            adm_generic_s, 1,1,1 },
-  { "dstate",            adm_generic_b, 1,1,1 },
+  { "cstate",            adm_generic_s, 1,1,0 },
+  { "dstate",            adm_generic_b, 1,1,0 },
   { "dump",              adm_dump,      1,1,1 },
   { "dump-xml",          adm_dump_xml,  1,1,1 },
   { "create-md",         adm_create_md, 1,1,0 },
@@ -230,10 +235,13 @@ struct adm_cmd cmds[] = {
   { "sh-md-dev",         sh_md_dev,     2,1,0 },
   { "sh-md-idx",         sh_md_idx,     2,1,0 },
   { "sh-ip",             sh_ip,         0,1,0 },
+  { "sh-b-pri",          sh_b_pri,      2,1,0 },
   { "pri-on-incon-degr", adm_khelper,   3,1,0 },
   { "pri-lost-after-sb", adm_khelper,   3,1,0 },
   { "outdate-peer",      adm_khelper,   3,1,0 },
   { "local-io-error",    adm_khelper,   3,1,0 },
+  { "pri-lost",          adm_khelper,   3,1,0 },
+  { "split-brain",       adm_khelper,   3,1,0 },
   { "set-gi",            admm_generic,  4,1,0 },
   { "suspend-io",        adm_generic_s, 4,1,0 },
   { "resume-io",         adm_generic_s, 4,1,0 },
@@ -537,6 +545,17 @@ static int sh_md_idx(struct d_resource* res,const char* unused __attribute((unus
   return 0;
 }
 
+static int sh_b_pri(struct d_resource *res,const char* unused __attribute((unused)))
+{
+  char* val;
+
+  val = get_opt_val(res->startup_options, "become-primary-on", NULL);
+  if ( val && ( !strcmp(val,nodeinfo.nodename) ||
+		!strcmp(val,"both") ) ) {
+    return adm_generic_s(res,"primary");
+  }
+  return 0;
+}
 
 static int sh_mod_parms(struct d_resource* res __attribute((unused)),const char* unused __attribute((unused)))
 {
@@ -606,9 +625,7 @@ static void expand_opts(struct d_option* co, struct d_option** opts)
   while(co) {
     if(!find_opt(*opts,co->name)) {
       // prepend new item to opts
-      no = malloc(sizeof(struct d_option));
-      no->name = strdup(co->name);
-      no->value = co->value ? strdup(co->value) : NULL ;
+      no = new_opt(strdup(co->name), co->value ? strdup(co->value) : NULL);
       no->next = *opts;
       *opts = no;
     }
@@ -656,7 +673,41 @@ static void alarm_handler(int __attribute((unused)) signo)
   alarm_raised=1;
 }
 
-pid_t m_system(char** argv,int flags)
+static inline const char* shell_escape(char* s)
+{
+	/* ugly static buffer. so what. */
+	static char buffer[1024];
+	char *c = buffer;
+
+	if (s == NULL)
+		return s;
+
+	while (1) {
+		if (buffer + sizeof(buffer) < c+2)
+			break;
+
+		switch(*s) {
+		case 0: /* terminator */
+			break;
+		/* set of 'clean' characters */
+		case '%': case '+': case '-': case '.': case '/':
+		case '0' ... '9':
+		case ':': case '=': case '@':
+		case 'A' ... 'Z':
+		case '_':
+		case 'a' ... 'z':
+			break;
+		/* escape everything else */
+		default:
+			*c++ = '\\';
+		}
+		*c++ = *s++;
+	}
+	*c = '\0';
+	return buffer;
+}
+
+pid_t m_system(char** argv, int flags, struct d_resource *res)
 {
   pid_t pid;
   int status,rv=-1;
@@ -670,8 +721,10 @@ pid_t m_system(char** argv,int flags)
   sa.sa_flags=0;
 
   if(dry_run || verbose) {
+    if (sh_varname && *cmdline)
+      fprintf(stdout,"%s=%s\n", sh_varname, shell_escape(res->name));
     while(*cmdline) {
-      fprintf(stdout,"%s ",*cmdline++);
+      fprintf(stdout,"%s ", shell_escape(*cmdline++));
     }
     fprintf(stdout,"\n");
     if (dry_run) return 0;
@@ -695,7 +748,7 @@ pid_t m_system(char** argv,int flags)
     alarm_raised=0;
     switch(flags & SLEEPS_MASK) {
     case SLEEPS_SHORT:     timeout = 5; break;
-    case SLEEPS_LONG:      timeout = 120; break;
+    case SLEEPS_LONG:      timeout = COMM_TIMEOUT+1; break;
     case SLEEPS_VERY_LONG: timeout = 600; break;
     default:
 	fprintf(stderr,"logic bug in %s:%d\n",__FILE__,__LINE__);
@@ -759,6 +812,19 @@ pid_t m_system(char** argv,int flags)
     OPT=OPT->next; \
   }
 
+#define make_options_wait(OPT) \
+  while(OPT) { \
+    if(!strcmp(OPT->name,"become-primary-on")) {\
+      OPT=OPT->next; continue; \
+    } \
+    if(OPT->value) { \
+      ssprintf(argv[NA(argc)],"--%s=%s",OPT->name,OPT->value); \
+    } else { \
+      ssprintf(argv[NA(argc)],"--%s",OPT->name); \
+    } \
+    OPT=OPT->next; \
+  }
+
 int adm_attach(struct d_resource* res,const char* unused __attribute((unused)))
 {
   char* argv[MAX_ARGS];
@@ -781,7 +847,7 @@ int adm_attach(struct d_resource* res,const char* unused __attribute((unused)))
   make_options(opt);
   argv[NA(argc)]=0;
 
-  return m_system(argv,SLEEPS_LONG);
+  return m_system(argv, SLEEPS_LONG, res);
 }
 
 struct d_option* find_opt(struct d_option* base,char* name)
@@ -811,7 +877,7 @@ int adm_resize(struct d_resource* res,const char* unused __attribute((unused)))
   }
   argv[NA(argc)]=0;
 
-  return m_system(argv,SLEEPS_SHORT);
+  return m_system(argv, SLEEPS_SHORT, res);
 }
 
 int _admm_generic(struct d_resource* res ,const char* cmd, int flags)
@@ -843,7 +909,7 @@ int _admm_generic(struct d_resource* res ,const char* cmd, int flags)
 
   argv[NA(argc)]=0;
 
-  return m_system(argv,flags);
+  return m_system(argv, flags, res);
 }
 
 static int admm_generic(struct d_resource* res ,const char* cmd)
@@ -864,7 +930,7 @@ static int adm_generic(struct d_resource* res,const char* cmd,int flags)
   }
   argv[NA(argc)]=0;
 
-  return m_system(argv,flags);
+  return m_system(argv, flags, res);
 }
 
 int adm_generic_s(struct d_resource* res,const char* cmd)
@@ -891,8 +957,6 @@ static int adm_generic_b(struct d_resource* res,const char* cmd)
   return rv;
 }
 
-static char* get_opt_val(struct d_option*,const char*,char*);
-
 static int adm_khelper(struct d_resource* res ,const char* cmd)
 {
   int rv=0;
@@ -904,7 +968,7 @@ static int adm_khelper(struct d_resource* res ,const char* cmd)
 
   if( (sh_cmd = get_opt_val(res->handlers,cmd,NULL)) ) {
     argv[2]=sh_cmd;
-    rv = m_system(argv,SLEEPS_VERY_LONG);
+    rv = m_system(argv, SLEEPS_VERY_LONG, res);
   }
   return rv;
 }
@@ -954,7 +1018,7 @@ int adm_connect(struct d_resource* res,const char* unused __attribute((unused)))
 
   argv[NA(argc)]=0;
 
-  return m_system(argv,SLEEPS_SHORT);
+  return m_system(argv, SLEEPS_SHORT, res);
 }
 
 struct d_resource* res_by_name(const char *name);
@@ -990,7 +1054,7 @@ int adm_syncer(struct d_resource* res,const char* unused __attribute((unused)))
   make_options(opt);
   argv[NA(argc)]=0;
 
-  return m_system(argv,SLEEPS_SHORT);
+  return m_system(argv, SLEEPS_SHORT, res);
 }
 
 static int adm_up(struct d_resource* res,const char* unused __attribute((unused)))
@@ -1012,10 +1076,10 @@ static int adm_wait_c(struct d_resource* res ,const char* unused __attribute((un
   argv[NA(argc)]=res->me->device;
   argv[NA(argc)]="wait-connect";
   opt=res->startup_options;
-  make_options(opt);
+  make_options_wait(opt);
   argv[NA(argc)]=0;
 
-  rv = m_system(argv,SLEEPS_FOREVER);
+  rv = m_system(argv, SLEEPS_FOREVER, res);
 
   return rv;
 }
@@ -1196,6 +1260,9 @@ static int adm_wait_ci(struct d_resource* ignored __attribute((unused)),const ch
   sigaction(SIGCHLD,&sa,&so);
 
   pids = alloca( nr_resources * sizeof(pid_t) );
+  /* alloca can not fail, it can "only" overflow the stack :)
+   * but it needs to be initialized anyways! */
+  memset(pids,0,nr_resources * sizeof(pid_t));
 
   for_each_resource(res,t,config) {
     argc=0;
@@ -1203,10 +1270,10 @@ static int adm_wait_ci(struct d_resource* ignored __attribute((unused)),const ch
     argv[NA(argc)]=res->me->device;
     argv[NA(argc)]="wait-connect";
     opt=res->startup_options;
-    make_options(opt);
+    make_options_wait(opt);
     argv[NA(argc)]=0;
 
-    pids[i++]=m_system(argv,RETURN_PID);
+    pids[i++]=m_system(argv, RETURN_PID, res);
   }
 
   wtime = global_options.dialog_refresh ?: -1;
@@ -1551,6 +1618,7 @@ void sanity_check_perm()
 void validate_resource(struct d_resource * res)
 {
   struct d_option* opt;
+  char *bpo;
 
   if (!res->protocol) {
     if (!common || !common->protocol) {
@@ -1607,6 +1675,16 @@ void validate_resource(struct d_resource * res)
   if ( (opt = find_opt(res->handlers, "outdate-peer")) ) {
     if(strstr(opt->value,"drbd-peer-outdater")) sanity_check_perm();
   }
+
+  bpo = get_opt_val(res->startup_options, "become-primary-on", "undef");
+  opt = find_opt(res->net_options, "allow-two-primaries");
+  if(!strcmp(bpo,"both") && opt == NULL) {
+    fprintf(stderr,
+	    "In resource %s:\n"
+	    "become-primary-on is set to both, but allow-two-primaries "
+	    "is not set.\n", res->name);
+    config_valid = 0;
+  }
 }
 
 static void global_validate(void)
@@ -1625,6 +1703,8 @@ int main(int argc, char** argv)
   struct adm_cmd *cmd;
   struct d_resource *res,*tmp;
   char *env_drbd_nodename = NULL;
+  int is_dump_xml;
+  int is_dump;
 
   drbdsetup=NULL;
   drbdmeta=NULL;
@@ -1723,6 +1803,29 @@ int main(int argc, char** argv)
 	    find_drbdcmd(&drbdmeta,pathes);
 	  }
 	  break;
+	case 'n':
+	  {
+	    char *c;
+            int shell_var_name_ok = 1;
+	    for (c = optarg; *c && shell_var_name_ok; c++) {
+	      switch(*c) {
+		case 'a' ... 'z':
+                case 'A' ... 'Z':
+                case '0' ... '9':
+                case '_':
+                  break;
+                default:
+                  shell_var_name_ok = 0;
+              }
+            }
+            if (shell_var_name_ok)
+              sh_varname = optarg;
+            else
+              fprintf(stderr,"ignored --sh-varname=%s: "
+                  "contains suspect characters, allowed set is [a-zA-Z0-9_]\n",
+                  optarg);
+	  }
+	  break;
 	case '?':
 	  // commented out, since opterr=1
 	  //fprintf(stderr,"Unknown option %s\n",argv[optind-1]);
@@ -1835,14 +1938,15 @@ int main(int argc, char** argv)
 
   uc_node(global_options.usage_count);
 
+  is_dump_xml = (cmd->function == adm_dump_xml);
+  is_dump = (is_dump_xml || cmd->function == adm_dump);
+  if (!is_dump || dry_run) expand_common();
+
   if(cmd->res_name_required)
     {
-      int is_dump_xml = (cmd->function == adm_dump_xml);
-      int is_dump = (is_dump_xml || cmd->function == adm_dump);
       if (optind + 1 > argc && !is_dump)
         print_usage_and_exit("missing arguments"); // arguments missing.
 
-      if (!is_dump || dry_run) expand_common();
       global_validate();
       if (!is_dump) {
 	if(!config_valid) exit(E_config_invalid);
@@ -1895,7 +1999,7 @@ int main(int argc, char** argv)
       }
     }
 
-  run_dcmds();
+  rv |= run_dcmds();
 
   free_config(config);
 
