@@ -3,9 +3,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2006-2007, LINBIT Information Technologies GmbH
-   Copyright (C) 2006-2007, Philipp Reisner <philipp.reisner@linbit.com>
-   Copyright (C) 2006-2007, Lars Ellenberg  <lars.ellenberg@linbit.com>
+   Copyright (C) 2006-2008, LINBIT Information Technologies GmbH
+   Copyright (C) 2006-2008, Philipp Reisner <philipp.reisner@linbit.com>
+   Copyright (C) 2006-2008, Lars Ellenberg  <lars.ellenberg@linbit.com>
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -150,7 +150,7 @@ void range_check(const enum range_checks what, const char *name,
 	}
 }
 
-static struct d_option *new_opt(char *name, char *value)
+struct d_option *new_opt(char *name, char *value)
 {
 	struct d_option *cn = malloc(sizeof(struct d_option));
 
@@ -561,7 +561,7 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 			goto exit_loop;
 		default:
 			pe_expected_got("protocol | on | disk | net | syncer |"
-					" startup | handler",token);
+					" startup | handlers",token);
 		}
 	}
 
