@@ -3,9 +3,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2001-2007, LINBIT Information Technologies GmbH.
-   Copyright (C) 1999-2007, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
+   Copyright (C) 2001-2008, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2008, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2002-2008, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -124,14 +124,15 @@ struct drbd_cmd {
 		} gp; // for generic_get_cmd, get_usage
 		struct {
 			struct option *options;
-			int (*proc_event)();
+			int (*proc_event)(unsigned int, int, 
+					  struct drbd_nl_cfg_reply *);
 		} ep; // for events_cmd, events_usage
 	};
 };
 
 
 // Connector functions
-#define NL_TIME 10000
+#define NL_TIME (COMM_TIMEOUT*1000)
 int open_cn();
 int send_cn(int sk_nl, struct nlmsghdr* nl_hdr, int size);
 int receive_cn(int sk_nl, struct nlmsghdr* nl_hdr, int size, int timeout_ms);
@@ -191,9 +192,9 @@ void show_bit(struct drbd_option *od, unsigned short* tp);
 void show_string(struct drbd_option *od, unsigned short* tp);
 
 // sub functions for events_cmd
-int print_state(unsigned int seq, struct drbd_nl_cfg_reply *reply);
-int w_connected_state(unsigned int seq, struct drbd_nl_cfg_reply *reply);
-int w_synced_state(unsigned int seq, struct drbd_nl_cfg_reply *reply);
+int print_state(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
+int w_connected_state(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
+int w_synced_state(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
 
 const char *on_error[] = {
 	[PassOn]         = "pass_on",
@@ -240,6 +241,7 @@ const char *rrcf_n[] = {
 struct option wait_cmds_options[] = {
 	{ "wfc-timeout",required_argument, 0, 't' },
 	{ "degr-wfc-timeout",required_argument,0,'d'},
+	{ "wait-after-sb",no_argument,0,'w'},
 	{ 0,            0,           0,  0  }
 };
 
@@ -393,7 +395,8 @@ static const char *error_messages[] = {
 	EM(PauseFlagAlreadyClear) = "Sync-pause flag is already cleared",
 	EM(DiskLowerThanOutdated) = "Disk state is lower than outdated",
 	EM(HaveNoDiskConfig) = "Device does not have a disk-config",
-	EM(ProtocolCRequired) = "Protocol C required"
+	EM(ProtocolCRequired) = "Protocol C required",
+	EM(VMallocFailed) = "vmalloc() failed. Out of memory?"
 };
 #define MAX_ERROR (sizeof(error_messages)/sizeof(*error_messages))
 const char * error_to_string(int err_no)
@@ -475,6 +478,15 @@ struct drbd_tag_list *create_tag_list(int size)
 
 void add_tag(struct drbd_tag_list *tl, int tag, void *data, int data_len)
 {
+	if(data_len > tag_descriptions[tag_number(tag)].max_len) {
+		fprintf(stderr, "The value for %s may only be %d byte long."
+			" You requested %d.\n",
+			tag_descriptions[tag_number(tag)].name,
+			tag_descriptions[tag_number(tag)].max_len,
+			data_len);
+		exit(20);
+	}
+
 	if( (tl->tag_list_cpos - tl->tag_list_start) + data_len
 	    > tl->tag_size ) {
 		fprintf(stderr, "Tag list size exceeded!\n");
@@ -733,7 +745,7 @@ int print_config_error( struct drbd_nl_cfg_reply *reply)
 	if (err_no == SS_Success) return rv;
 
 	if ( ( err_no >= AfterLastRetCode || err_no <= RetCodeBase ) &&
-	     ( err_no > SS_CW_NoNeed || err_no < SS_DeviceInUse) ) {
+	     ( err_no > SS_CW_NoNeed || err_no < SS_LowerThanOutdated) ) {
 		fprintf(stderr,"Error code %d unknown.\n"
 			"You should updated the drbd userland tools.\n",err_no);
 		rv = 20;
@@ -806,6 +818,7 @@ int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 
 	if(rv == 0) {
 		//dump_tag_list(tl->tag_list_start);
+		int received;
 		sk_nl = open_cn();
 		if(sk_nl < 0) return 20;
 
@@ -813,13 +826,15 @@ int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 		tl->drbd_p_header->drbd_minor = minor;
 		tl->drbd_p_header->flags = flags;
 
-		call_drbd(sk_nl,tl, (struct nlmsghdr*)buffer,RCV_SIZE,NL_TIME);
+		received = call_drbd(sk_nl,tl, (struct nlmsghdr*)buffer,RCV_SIZE,NL_TIME);
 
 		close_cn(sk_nl);
 
-		reply = (struct drbd_nl_cfg_reply *)
-			((struct cn_msg *)NLMSG_DATA(buffer))->data;
-		rv = print_config_error(reply);
+		if (received >= 0) {
+			reply = (struct drbd_nl_cfg_reply *)
+				((struct cn_msg *)NLMSG_DATA(buffer))->data;
+			rv = print_config_error(reply);
+		}
 	}
 	free_tag_list(tl);
 
@@ -1004,6 +1019,9 @@ int generic_get_cmd(struct drbd_cmd *cm, int minor, int argc,
 	reply = (struct drbd_nl_cfg_reply *)
 		((struct cn_msg *)NLMSG_DATA(buffer))->data;
 
+	if (reply->ret_code != NoError)
+		return print_config_error(reply);
+
 	rv = cm->gp.show_function(cm,minor,reply->tag_list);
 
 	if(dump_tag_list(reply->tag_list)) {
@@ -1178,7 +1196,8 @@ int down_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 	return rv;
 }
 
-int print_state(unsigned int seq, struct drbd_nl_cfg_reply *reply)
+int print_state(unsigned int seq, int u __attribute((unused)),
+		struct drbd_nl_cfg_reply *reply)
 {
 	drbd_state_t state;
 	char* str;
@@ -1230,6 +1249,7 @@ int print_state(unsigned int seq, struct drbd_nl_cfg_reply *reply)
 }
 
 int w_connected_state(unsigned int seq __attribute((unused)),
+		      int wait_after_sb,
 		      struct drbd_nl_cfg_reply *reply)
 {
 	drbd_state_t state;
@@ -1237,6 +1257,7 @@ int w_connected_state(unsigned int seq __attribute((unused)),
 	if(reply->packet_type == P_get_state) {
 		if(consume_tag_int(T_state_i,reply->tag_list,(int*)&state.i)) {
 			if(state.conn >= Connected) return 0;
+			if(!wait_after_sb && state.conn < Unconnected) return 0;
 		} else fprintf(stderr,"Missing tag !?\n");
 	}
 
@@ -1244,14 +1265,15 @@ int w_connected_state(unsigned int seq __attribute((unused)),
 }
 
 int w_synced_state(unsigned int seq __attribute((unused)),
+		   int wait_after_sb,
 		   struct drbd_nl_cfg_reply *reply)
 {
 	drbd_state_t state;
 
 	if(reply->packet_type == P_get_state) {
 		if(consume_tag_int(T_state_i,reply->tag_list,(int*)&state.i)) {
-			if(state.conn == Connected || state.conn < Unconnected )
-				return 0;
+			if(state.conn == Connected) return 0;
+			if(!wait_after_sb && state.conn < Unconnected) return 0;
 		} else fprintf(stderr,"Missing tag !?\n");
 	}
 	return 1;
@@ -1264,11 +1286,12 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 	struct drbd_nl_cfg_reply *reply;
 	struct drbd_tag_list *tl;
 	struct option *lo;
-	unsigned int seq=0;
+	unsigned int b_seq=0, r_seq=0;
 	int sk_nl,c,cont=1,rr,i,last;
 	int unfiltered=0, all_devices=0;
 	int wfc_timeout=0, degr_wfc_timeout=0,timeout_ms;
 	struct timeval before,after;
+	int wasb=0;
 
 	lo = cm->ep.options;
 
@@ -1298,6 +1321,9 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 				return 20;
 			}
 			break;
+		case 'w':
+			wasb=1;
+			break;
 		}
 	}
 
@@ -1359,11 +1385,20 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 
 		// dump_tag_list(reply->tag_list);
 
-		if(!unfiltered && cn_reply->seq <= seq) continue;
-		seq = cn_reply->seq;
+		/* There are two value spaces for sequence numbers. The first
+		   is the one created by this drbdsetup instance, the kernel's
+		   reply packets simply echo those sequence numbers.
+		   The second is created by the kernel's broadcast packets. */
+		if(cn_reply->ack==0) { // broadcasts
+			if(!unfiltered && cn_reply->seq <= b_seq) continue;
+			b_seq = cn_reply->seq;
+		} else { // replies to drbdsetup packes
+			if(!unfiltered && cn_reply->seq <= r_seq) continue;
+			r_seq = cn_reply->seq;
+		}
 
 		if( all_devices || minor == reply->minor ) {
-			cont=cm->ep.proc_event(cn_reply->seq, reply);
+			cont=cm->ep.proc_event(cn_reply->seq, wasb, reply);
 		}
 	} while(cont);
 
@@ -1624,8 +1659,13 @@ int open_cn()
 	if (err == -1) {
 		err = errno;
 		perror("bind() failed");
-		if(err == ENOENT) {
-			fprintf(stderr,"DRBD driver not present in the kernel?\n");
+		switch(err) {
+		case ENOENT:
+			fprintf(stderr,"Connector module not loaded? Try 'modprobe cn'.\n");
+			break;
+		case EPERM:
+			fprintf(stderr,"Missing privileges? You should run this as root.\n");
+			break;
 		}
 		return -1;
 	}
@@ -1741,6 +1781,32 @@ void close_cn(int sk_nl)
 	close(sk_nl);
 }
 
+void ensure_drbd_driver_is_present(void)
+{
+	struct drbd_tag_list *tl;
+	char buffer[4096];
+	int sk_nl, rr;
+
+	sk_nl = open_cn();
+	/* Might print:
+	   Missing privileges? You should run this as root.
+	   Connector module not loaded? try 'modprobe cn'. */
+	if (sk_nl < 0) exit(20);
+
+	tl = create_tag_list(2);
+	add_tag(tl, TT_END, NULL, 0); // close the tag list
+
+	tl->drbd_p_header->packet_type = P_get_state;
+	tl->drbd_p_header->drbd_minor = 0;
+	tl->drbd_p_header->flags = 0;
+
+	rr = call_drbd(sk_nl, tl, (struct nlmsghdr*)buffer, 4096, 500);
+	/* Might print: (after 500ms)
+	   No response from the DRBD driver! Is the module loaded? */
+	close_cn(sk_nl);
+	if (rr == -2) exit(20);
+}
+
 int main(int argc, char** argv)
 {
 	int minor;
@@ -1777,6 +1843,8 @@ int main(int argc, char** argv)
 
 	cmd=find_cmd_by_name(argv[2]);
 
+	ensure_drbd_driver_is_present();
+
 	if(cmd) {
 		lock_fd = dt_lock_drbd(argv[1]);
 		minor=dt_minor_of_dev(argv[1]);
